#include "../include/vmi_simplifier.h"

namespace VMI 
{
GLMmodel *pmodel = NULL; // The .obj model

GLubyte *pixels = NULL;

Camera *cameras = NULL;
GLuint numCameras = 0;

Color *colors = NULL;

GLuint **histogram = NULL,
        *queries = NULL;

Mesh *mesh = NULL;

GLdouble *initialIs = NULL;

GLsizei width  = 256,
        height = 256;

GLboolean bEnableOffScreen = GL_TRUE,
          bBeQuiet         = GL_FALSE,
          bSaveLog         = GL_FALSE,
          bLoadCamerasFromFile = GL_FALSE;

GLuint cameraType = 0,
       numDemandedTriangles = 0;

GLdouble radius = 1.3, fov = 60.0;

char filename[MAX_CHAR] = {'\0'};

GLuint fb, depth_rb, color_tex;

int vmiWin = 0;

//	For progress update.
Geometry::TIPOFUNC	mUPB;

}

using namespace	VMI;

void VMI::init(void)
{
    // Clear color
    glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
    
    // Initialize viewing values
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(fov, (GLdouble)width / height, 0.1, 40.0);
    
    glPixelStorei(GL_PACK_ALIGNMENT, 1);
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    
    glShadeModel(GL_FLAT);
    
    glDisable(GL_BLEND);

    //glEnable(GL_CULL_FACE); // important

    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LEQUAL);

    if (GL_TRUE != glewIsSupported((const char*) "GL_EXT_framebuffer_object"))
    {
        fprintf(stderr,"GL_EXT_framebuffer_object extension is not available!\n");
        exit(1);
    } else {
         // create objects
        glGenFramebuffersEXT(1, &fb);        // frame buffer
        glGenTextures(1, &color_tex);        // texture
        glGenRenderbuffersEXT(1, &depth_rb); // render buffer
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb);

        // initialize color texture
        glBindTexture(GL_TEXTURE_2D, color_tex);
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0,
                     GL_RGBA, GL_UNSIGNED_BYTE, NULL);
        // texture parameters
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

        // attach texture to framebuffer color buffer
        glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,
                                  GL_COLOR_ATTACHMENT0_EXT,
                                  GL_TEXTURE_2D, color_tex, 0);
        // initialize depth renderbuffer
        glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, depth_rb);
        glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT,
                                 GL_DEPTH_COMPONENT24, width, height);
        // attach renderbuffer to framebuffer depth buffer
        glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT,
                                     GL_DEPTH_ATTACHMENT_EXT,
                                     GL_RENDERBUFFER_EXT, depth_rb);

        // Check framebuffer completeness at the end of initialization.
        CHECK_FRAMEBUFFER_STATUS();
    }
    /*if (GL_TRUE != glewIsSupported((const char*) "GL_ARB_fragment_shader"))
    {
        fprintf(stderr,"GL_ARB_fragment_shader extension is not available!\n");
    }
    if (GL_TRUE != glewIsSupported((const char*) "GL_ARB_vertex_shader"))
    {
        fprintf(stderr,"GL_ARB_vertex_shader extension is not available!\n");
    }
    if (GL_TRUE != glewIsSupported((const char*) "GL_ARB_shader_objects"))
    {
        fprintf(stderr,"GL_ARB_shader_objects extension is not available!\n");
    }*/
    if (GL_TRUE != glewIsSupported((const char*) "GL_ARB_vertex_buffer_object"))
    {
        fprintf(stderr,"GL_ARB_vertex_buffer_object extension is not available!\n");
    }
    if (GL_TRUE != glewIsSupported((const char*) "GL_ARB_occlusion_query"))
    {
        fprintf(stderr,"GL_ARB_occlusion_query extension is not available!\n");
    } else {
#ifdef USE_OCCLUSION_QUERY
        // Generate a list of occlusion queries
        queries = (GLuint *)malloc(mesh->currentNumTriangles * sizeof(GLuint));
        glGenQueriesARB(mesh->currentNumTriangles, queries);y
#endif
    }
    if (GL_TRUE != glewIsSupported((const char*) "GL_ARB_imaging"))
    {
        fprintf(stderr,"GL_ARB_imaging extension is not available!\n");
    }
    if (GL_TRUE != glewIsSupported((const char*) "GL_ARB_multitexture"))
    {
        fprintf(stderr,"GL_ARB_multitexture extension is not available!\n");
    }
#ifdef _WIN32
    if (GL_TRUE != wglewIsSupported((const char*) "WGL_EXT_swap_control"))
    {
        fprintf(stderr,"WGL_EXT_swap_control extension is not available!\n");
    }
#endif
    
    // Allocate memory for the image pixels
    pixels = (GLubyte *)malloc(width * height * 4 * sizeof(GLubyte));
    
    // Allocate memory for colors
    colors = initColors(mesh->currentNumTriangles);

    // Set a different color for every triangle 
    setColors4(colors, mesh->currentNumTriangles);
}

void VMI::applyHWAcceleration(void) {
    
#ifdef VERTEX_ARRAY_INTERLEAVE
    setupInterleave(mesh, colors);
#endif
#ifdef VERTEX_ARRAY
    setupVertexArray(mesh, colors);
#endif
#ifdef VERTEX_BUFFER_OBJECTS
    setupVertexBufferObjects(mesh, colors);
#endif

}

void VMI::setOrthographicProjection(void) {
    // Switch to projection mode
    glMatrixMode(GL_PROJECTION);
    // Save previous matrix which contains the 
    // settings for the perspective projection
    glPushMatrix();
    // Reset matrix
    glLoadIdentity();
    glOrtho(0.0, (GLdouble)HISTOGRAM_SIZE, 0.0, (GLdouble)height, -1.0, 1.0);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
}

void VMI::resetPerspectiveProjection(void) {
    // Set the current matrix to GL_PROJECTION
    glMatrixMode(GL_PROJECTION);
    // Restore previous settings
    glPopMatrix();
    // Get back to GL_MODELVIEW matrix
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
}

void VMI::renderBitmapString(float x, float y, void *font, char *string) {
    char *c;
    // Set position to start drawing fonts
    glRasterPos2f(x, y);
    // Loop all the characters in the string
    for (c=string; *c != '\0'; c++)
        glutBitmapCharacter(font, *c);
}

void VMI::renderGeometry(void) {
#ifdef IMMEDIATE_MODE
    GLuint i, v1, v2, v3;

    // Immediate mode
    glBegin (GL_TRIANGLES);
    for (i=0; i<mesh->numTriangles; i++) {
        if (mesh->triangles[i].enable == TRUE) {
            
            v1= mesh->triangles[i].indices[0];
            v2= mesh->triangles[i].indices[1];
            v3= mesh->triangles[i].indices[2];
            
            glColor4ub(colors[i].r, colors[i].g, colors[i].b, colors[i].a);
            glVertex3f(mesh->vertices[v1].x, mesh->vertices[v1].y, mesh->vertices[v1].z);
            glVertex3f(mesh->vertices[v2].x, mesh->vertices[v2].y, mesh->vertices[v2].z);
            glVertex3f(mesh->vertices[v3].x, mesh->vertices[v3].y, mesh->vertices[v3].z);
        }
    }
    glEnd();
#else
#ifdef VERTEX_ARRAY
    glDrawArrays(GL_TRIANGLES, 0, mesh->currentNumTriangles * 3);
#endif
#ifdef VERTEX_ARRAY_INTERLEAVE
    glInterleavedArrays (GL_T2F_C4UB_V3F, 0, intertwined);
    glDrawArrays(GL_TRIANGLES, 0, mesh->currentNumTriangles * 3);
#endif
#ifdef VERTEX_BUFFER_OBJECTS
    glDrawArrays(GL_TRIANGLES, 0, mesh->currentNumTriangles * 3);
#endif

#endif
}
void VMI::drawTexture(void) {

    // Clear color and depth buffers
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // Switch to projection mode
    glMatrixMode(GL_PROJECTION);
    // Save previous matrix which contains the 
    // settings for the perspective projection
    glPushMatrix();
    // Reset matrix
    glLoadIdentity();
    glOrtho(0.0, (GLdouble)width, 0.0, (GLdouble)height, 1.0, 1.0);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, color_tex);
    glBegin(GL_QUADS);
    glTexCoord2f(0.0, 0.0); glVertex3f(-1.0, -1.0, 0.0);
    glTexCoord2f(0.0, 1.0); glVertex3f(-1.0, 1.0, 0.0);
    glTexCoord2f(1.0, 1.0); glVertex3f( 1.0, 1.0, 0.0);
    glTexCoord2f(1.0, 0.0); glVertex3f( 1.0, -1.0, 0.0);
    glEnd();
    glFlush();
    glDisable(GL_TEXTURE_2D);

    resetPerspectiveProjection();

    glutSwapBuffers();
}
void VMI::renderScene(GLuint **histogram, GLuint numCameras)
{
    GLuint i;

    // draw to the frame buffer
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb);
        
    // Apply HW acceleration OpenGL Technique
    applyHWAcceleration(); 

    glDrawBuffer(GL_BACK);
    glReadBuffer(GL_BACK);

    // Get the projected areas for all cameras
    for (i=0; i<numCameras; i++) {
        
        // Clear color and depth buffers
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        
        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();
        
        // Camera  i
        gluLookAt(cameras[i].eyeX, cameras[i].eyeY, cameras[i].eyeZ,
                  cameras[i].centerX, cameras[i].centerY, cameras[i].centerZ,
                  cameras[i].upX, cameras[i].upY, cameras[i].upZ);
        
        renderGeometry();
        glFlush();
        
        ///////////////////////////////////////////////////////////////////////////
        
        resetSWHistogram(histogram[i], mesh->numTriangles);
        
#ifdef USE_OCCLUSION_QUERY
        getSWHistoByOcclusionQuery(mesh, colors, histogram[i]);
#else // HYBRID_HW_SW_HISTOGRAM
        getSWHistogram(histogram[i], pixels);
#endif
    }

    // draw to the window, reading from the color texture
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);

    if (!bEnableOffScreen) drawTexture();
    
    ///////////////////////////////////////////////////////////////////////////
}

void VMI::freeMemory(void) {

	  // Free memory
    deleteMesh(mesh);

    //glmDelete(pmodel);

    free(pixels);

    free(cameras);

    free(colors);
    
    deleteHistogram(histogram, numCameras);

    free(initialIs);
    
    glDeleteRenderbuffersEXT(1, &depth_rb);
    
    glDeleteRenderbuffersEXT(1, &color_tex);
    
    glDeleteFramebuffersEXT(1, &fb);
#ifdef USE_OCCLUSION_QUERY
    glDeleteQueriesARB(mesh->currentNumTriangles, queries);
#endif

		glutDestroyWindow(vmiWin);
}

void VMI::display(void)
{
	clock_t start, finish;
	GLdouble timeElapsed = 0.0;
	char s[MAX_CHAR];

	// Clear color and depth buffers
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glutSwapBuffers();

	start = clock();

	getProjectedAreas(histogram, numCameras);
	//printFullHistogram(histogram, numCameras, currentNumTriangles);

#ifdef SALIENCY
	printf("\nComputing Saliency Map...");
	computeSaliency(mesh, histogram, numCameras);
	printf("Ok\n");
	//getchar();
	//exit(1);
#endif

#ifdef KL // Kullback-Leibler
	printf("\nComputing viewpoint KL divergences...\n");
#endif
#ifdef MI // Mutual Information
	printf("\nComputing viewpoint Mutual Informations...\n");
#endif
#ifdef HE // Hellinger
	printf("\nComputing viewpoint HE divergences...\n");
#endif
#ifdef CS // Chi-Square
	printf("\nComputing viewpoint CS divergences...\n");
#endif

	computeCameraIs(histogram, numCameras, initialIs);

	printf("Ok\n");

	simplifyModel(mesh, numDemandedTriangles);

	finish = clock();
	timeElapsed = (GLdouble)(finish - start)/ CLOCKS_PER_SEC;
	printf("Time: %f\n", timeElapsed);

	glFlush();

#ifdef SALIENCY
#ifdef KL // Kullback-Leibler
	sprintf(s,"%s_%d_%d_VKL_S.obj", filename, numCameras, mesh->currentNumTriangles);
#endif
#ifdef MI // Mutual Information
	sprintf(s,"%s_%d_%d_VMI_S.obj", filename, numCameras, mesh->currentNumTriangles);
#endif
#ifdef HE // Hellinger
	sprintf(s,"%s_%d_%d_VHE_S.obj", filename, numCameras, mesh->currentNumTriangles);
#endif
#ifdef CS // Chi-Square
	sprintf(s,"%s_%d_%d_VCS_S.obj", filename, numCameras, mesh->currentNumTriangles);
#endif
#else
#ifdef KL // Kullback-Leibler
	sprintf(s,"%s_%d_%d_VKL.obj", filename, numCameras, mesh->currentNumTriangles);
#endif
#ifdef MI // Mutual Information
	sprintf(s,"%s_%d_%d_VMI.obj", filename, numCameras, mesh->currentNumTriangles);
#endif
#ifdef HE // Hellinger
	sprintf(s,"%s_%d_%d_VHE.obj", filename, numCameras, mesh->currentNumTriangles);
#endif
#ifdef CS // Chi-Square
	sprintf(s,"%s_%d_%d_VCS.obj", filename, numCameras, mesh->currentNumTriangles);
#endif
#endif

	//saveModel(s, pmodel, mesh);
}

void reshape(int w, int h)
{
	if (h != 0)
	{
		glViewport(0, 0, w, h);      // Set the viewport
		glMatrixMode(GL_PROJECTION); // Select the projection matrix
		glLoadIdentity();            // Reset The Projection Matrix
		//printf("Width: %d Height: %d\n", w, h);
		gluPerspective(fov, (GLdouble)w/h, 0.1, 40.0);
		//width = w;
		//height = h;
		glMatrixMode(GL_MODELVIEW);  // Switch back to the modelview matrix
		glLoadIdentity();
	}
}

void keyboard(unsigned char key, int x, int y)
{
	switch (key)
	{
		case 27:
			exit(0);
		default:
			break;
	}

	glutPostRedisplay();
}

/*  Main Loop
 *  Open window with initial window size, title bar, 
 *  RGBA display mode, and handle input events.
 */
//int main(int argc, char** argv)
//{
//    char s[MAX_CHAR];
//
//    process_cmdline(argc, argv);
//    
//    // Load a .obj model
//    pmodel = glmReadOBJ(argv[argc-1]);
//    if (!pmodel) exit(0);
//    glmUnitize(pmodel);
//    //glmFacetNormals(pmodel);
//    //glmVertexNormals(pmodel, 90.0);
//
//    mesh = initMesh(pmodel);
//    //printMesh(mesh);
//    //getchar();
//
//    if ((numDemandedTriangles == 0) || (numDemandedTriangles >= mesh->currentNumTriangles))
//        usage_error("Illegal number of triangles.");
//
//    printf("w: %d h: %d\n", width, height);
//    printf("t: %d c: %d o: %d r: %f\n", numDemandedTriangles, cameraType, bEnableOffScreen, radius);
//    //getchar();
//
//    // Get a filename without extension
//    strncpy(filename, argv[argc - 1], strlen(argv[argc - 1]) - 4);
//    
//    glutInit(&argc, argv);
//    glutInitDisplayMode(GLUT_DEPTH | GLUT_RGBA | GLUT_DOUBLE | GLUT_ALPHA); /* RGB and Alpha */
//    glutInitWindowSize(width, height);
//    glutInitWindowPosition(100, 100);
//
//#ifdef KL // Kullback-Leibler
//    sprintf(s, "VKL - [%s]", argv[argc-1]); 
//#endif
//#ifdef MI // Mutual Information
//    sprintf(s, "VMI - [%s]", argv[argc-1]); 
//#endif
//#ifdef HE // Hellinger
//    sprintf(s, "VHE - [%s]", argv[argc-1]); 
//#endif
//#ifdef CS // Chi-Square
//    sprintf(s, "VCS - [%s]", argv[argc-1]); 
//#endif
//    
//    vmiWin = glutCreateWindow(s);
//
//    glewInit();
//
//    init();
//
//    if (bLoadCamerasFromFile == GL_FALSE) {
//        switch (cameraType) {
//        case 0:
//            cameras = setCameras(radius, OCTAHEDRON, &numCameras);
//            printf("Number of cameras: %d\n", OCTAHEDRON);
//            break;
//        case 1:
//            cameras = setCameras(radius, ICOSAHEDRON, &numCameras);
//            printf("Number of cameras: %d\n", ICOSAHEDRON);
//            break;
//        case 2:
//            cameras = setCameras(radius, DODECAHEDRON, &numCameras);
//            printf("Number of cameras: %d\n", DODECAHEDRON);
//            break;
//        default:
//            break;
//        }
//    } else {
//        sprintf(s,"%s.cam", filename);
//        
//        cameras = loadCameras(radius, s, &numCameras);
//        //getchar();
//    }
//
//		histogram = initHistogram(mesh->currentNumTriangles, numCameras);
//
//    initialIs = initIs(numCameras);
//    
//		if (!bEnableOffScreen) {
//			  glutReshapeFunc(reshape);
//        glutKeyboardFunc(keyboard);
//			  glutDisplayFunc(display);
//		}
//		else {
//        display();
//    		freeMemory(); 		
//    }
//    exit(0);
//
//    glutMainLoop();
//    return 0;
//}
